#include <cstdio>
#include <string>

#include "msg.hh"
#include "settings.hh"

using std::string;
using std::FILE;
using std::sprintf;

#define BGR24 0x42475218  // BGR24 fourcc
#define BGR16 0x42475210  // BGR16 fourcc
#define BGR15 0x4247520F  // BGR15 fourcc

static const string GetInt(unsigned n)
{
    char Buf[64];
    sprintf(Buf, "%u", n);
    return Buf;
}

FILE *CreateWritingPipe(const char *CmdLine)
{
    /* NOTE: THIS REQUIRES "bufferprog" INSTALLED */
    string command = "bufferprog | nice -1 ";
    command += CmdLine;
    return popen(command.c_str(), "w");
}

static FILE* avifp = NULL;

void OpenAVI()
{
    static unsigned avinumber = 0;
    char CmdLine[512];

    string avifn = OUTPUT_FILE;
    
    string options = LAVCOPTS;

    sprintf(CmdLine,
        "mencoder - -o %s -ovc lavc -lavcopts %s:vbitrate=%u",
//      "  >& /dev/null",
        avifn.c_str(),
        options.c_str(),
        BITRATE
           );

    //sprintf(CmdLine, "cat >/dev/null");

    MessageCreatingFile(avifn.c_str());

    avifp = CreateWritingPipe(CmdLine);
}

void CloseAVI()
{
    pclose(avifp);
}

#define u32(n) (n)&255,((n)>>8)&255,((n)>>16)&255,((n)>>24)&255
#define u16(n) (n)&255,((n)>>8)&255
#define s4(s) s[0],s[1],s[2],s[3]

void SendFrame(const unsigned char *vidbuf,
               unsigned width,
               unsigned height,
               bool first_frame)
{
    const unsigned fourcc = BGR16;
    const unsigned framesize = width*height*2;
    
    if(first_frame)
    {
        const unsigned nframes = 0; //unknown
        const unsigned scale = 1000000;
        const unsigned scaled_fps = (unsigned)(scale * FPS);
        
        const unsigned SIZE_strh_vids = 4 + 4*2 + 2*2 + 8*4 + 2*4;
        const unsigned SIZE_strf_vids = 4*3 + 2*2 + 4*6;
        const unsigned SIZE_strl = 4+ 4+(4+SIZE_strh_vids) + 4+(4+SIZE_strf_vids);
        const unsigned SIZE_avih = 4*12;
        const unsigned SIZE_hdrl = 4+4+ (4+SIZE_avih) + 4 + (4+SIZE_strl);
        const unsigned SIZE_movi = 4 + nframes*(4+4+framesize);
        const unsigned SIZE_avi = 4+4+ (4+SIZE_hdrl) + 4 + (4+SIZE_movi);
        
        const unsigned char AVIheader[] =
        {
            s4("RIFF"),
            u32(SIZE_avi),
            s4("AVI "),
            s4("LIST"),
            u32(SIZE_hdrl),
             s4("hdrl"),
             
             s4("avih"),
             u32(SIZE_avih),
              u32(0),
              u32(0),
              u32(0),
              u32(0),
              u32(nframes),
              u32(0),
              u32(1), // one stream
              u32(0),
              u32(0),
              u32(0),
              u32(0),
              u32(0),
             
             s4("LIST"),
             
             u32(SIZE_strl),
              s4("strl"),
              
               s4("strh"),
               u32(SIZE_strh_vids),
                s4("vids"),
                u32(0),
                u32(0),
                u16(0),
                u16(0),
                u32(0),
                u32(scale),
                u32(scaled_fps),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
                u16(0),
                u16(0),
                u16(0),
                u16(0),
               
               s4("strf"),
               u32(SIZE_strf_vids),
                u32(0),
                u32(width),
                u32(height),
                u16(0),
                u16(0),
                u32(fourcc),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
                u32(0),
            
            s4("LIST"),
            u32(SIZE_movi),
             s4("movi")
        };
        
        fwrite(AVIheader, 1, sizeof(AVIheader), avifp);
    }

    const unsigned char header[] = { s4("00dc"), u32(framesize) };
    
    fwrite(header, 1, sizeof(header), avifp);
    fwrite(vidbuf, 1, framesize, avifp);
}
