/**************

 XPATCH
  Example of how this program is to be used:
  
     rm -rf /tmp/.wine-*
     LD_LIBRARY_PATH=/usr/lib/wine/lib \
     LD_PRELOAD=~bisqwit/nes/xgetimage/xpatch.so \
     /usr/lib/wine/bin/wine-pthread fam/FAMBLIP.EXE tinytoon.nes
  
  It logs video. Requires MEncoder.

 If USE_MEMFILE and USE_SHM are defined

   TODO: Document
 
 If USE_MEMFILE is defined but USE_SHM is not
   TODO: Document
 
 If USE_MEMFILE is not defined
   TODO: Document


****/

#include <dlfcn.h>
#include <X11/Xlib.h>
#include <fcntl.h>
#include <unistd.h>

#include <cstdio>
#include <algorithm>

#include "settings.hh"
#include "avifun.hh"

#define CALL_OLD

#ifdef USE_MEMFILE
 #ifdef USE_SHM
  #include <sys/ipc.h>
  #include <sys/shm.h>
 #else
  #include <sys/mman.h>
 #endif
#else
 #include "avifun.hh"
#endif

using std::fprintf;

static bool Certainly(const void *);


#define XPutImage_f_params \
    Display*         display , \
    Drawable         d , \
    GC           gc ,    \
    XImage*      image , \
    int          src_x , \
    int          src_y , \
    int          dest_x , \
    int          dest_y , \
    unsigned int     width , \
    unsigned int     height 

int XPutImage(XPutImage_f_params)
{
    /* This object will load the _real_ XPutImage
     * symbol and enable calling it
     */
    static struct caller
    {
        int (*orig) (XPutImage_f_params);
    public:
        caller()
        {
            void *f = dlopen("libXext.so", RTLD_NOW);
            
            orig = (int(*)(XPutImage_f_params))dlsym(f, "XPutImage");
            
            dlclose(f);
        }
    } caller;
    
    static struct sender
    {
#ifndef USE_SHM
        int fd;
#endif
#ifdef USE_MEMFILE
        char *buffer;
        unsigned size;
#else
        bool first;
#endif
    public:
#ifdef USE_MEMFILE
        sender() : buffer(NULL), size(256*240*2)
#else
        sender() : first(true)
#endif
        {
            /* Prevent the subprocesses getting the same
             * measures as this victim program did
             */
            unsetenv("LD_PRELOAD");
            
#ifdef USE_MEMFILE
 #ifdef USE_SHM
            int fd;
 #endif
            fd = open(MEMORY_FILE.c_str(),
                      O_RDWR | O_CREAT | O_TRUNC, 0666);
            if(fd == -1)
            {
                perror(MEMORY_FILE.c_str());
            }
            
 #ifdef USE_SHM
            int shmid = shmget(IPC_PRIVATE, size, IPC_CREAT | 0666);
            if(shmid == -1)
            {
                perror("shmget");
            }
            char Buf[64];
            sprintf(Buf, "%d", shmid);
            write(fd, Buf, strlen(Buf));
            close(fd);
            
            buffer = (char*) (shmat(shmid, NULL, 0));
 #else
            ftruncate(fd, size);
            
            buffer = (char*) (mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0));
            if(!buffer || buffer == (char*)(MAP_FAILED))
            {
                perror("mmap");
            }
 #endif
#else
            OpenAVI();
#endif
        }
        ~sender()
        {
#ifdef USE_MEMFILE
 #ifdef USE_SHM
            shmdt(buffer);
 #else
            close(fd);
            munmap(buffer, size);
 #endif
            unlink(MEMORY_FILE.c_str());
#else
            CloseAVI();
#endif
        }
        void Send(const char *imagedata, unsigned width, unsigned height)
        {
#ifdef USE_MEMFILE
            std::copy(imagedata, imagedat + width*height*2,
                      buffer);
#else
            if(width != 256 || height < 224 || height > 240) return;
            SendFrame((const unsigned char*) imagedata, width, height, first);
            first = false;
#endif
        }
    } sender;
    
    if(!Certainly((const void *)d))
    {
        //return caller.orig(display,d,gc,image,src_x,src_y,dest_x,dest_y,width,height);
        return 0;
    }

    //fprintf(stderr, "Put(%u,%u) -> %p - %u\n", width,height, gc, gc->gid);
    
    sender.Send(image->data, width, height);

/*
    fprintf(stderr, "width=%u height=%u src=(%u,%u) dest=(%u,%u)\n",
       width,height,
       src_x,src_y,
       dest_x,dest_y);
*/
    

#ifdef CALL_OLD
    return caller.orig(display,d,gc,image,src_x,src_y,dest_x,dest_y,width,height);
#endif
    
    return 0;
}

#include <set>
static bool Certainly(const void *p)
{
    return true;
    
    typedef const void *ptrt;
    
    const unsigned N = 8;
    
    static ptrt history[N];
    static unsigned ptr=0;
    
    history[ptr] = p;
    ptr = (ptr+1) % N;

    unsigned n = 0;
    for(unsigned a=0; a<N; ++a)
        if(history[a] == p)
        {
            ++n;
            if(n > 2) return true;
        }
    
    return false;
}

#include "xpatch.inc"
